function gadget:GetInfo()
  return {
    name      = "wraparound map",
    desc      = "teleport units from one map edge to another",
    author    = "knorke",
    date      = "Nov 2010",
    license   = "GNU GPL, v2 or later or horse",
    layer     = 0,
    enabled   = true,
  }
end

local movecrtlunits = {}
local debug = false

if not gadgetHandler:IsSyncedCode() then
  return false -- no unsynced code
end

local doScramble = Spring.GetMapOptions().ma_scrambleteleports or "none" -- possible options: none (no scramble), reciprocal (a<->b matching), shuffle ( a->b, b->c, c->a )

local random = math.random
local Echo = Spring.Echo
local GetUnitsInCylinder = Spring.GetUnitsInCylinder
local ValidUnitID = Spring.ValidUnitID
local GetUnitDirection = Spring.GetUnitDirection
local GiveOrderToUnit = Spring.GiveOrderToUnit
local MarkerAddPoint = Spring.MarkerAddPoint
local GetUnitBasePosition = Spring.GetUnitBasePosition
local GetGroundHeight = Spring.GetGroundHeight
local GetUnitHealth = Spring.GetUnitHealth
local RevertHeightMap = Spring.RevertHeightMap
local SetUnitPosition = Spring.SetUnitPosition


local exitmoveaway = 500	--how much further away from the exit will units move away?
local scrambleinterval =  5 * 60 * 30 -- 5 minutes
local teleportradius = 150
local teleportmove = teleportradius/3*2
local teleportradiussquared = teleportradius*teleportradius

local teleport = {}
teleport[1] = {}
teleport[1].x1=3733		--4 corners to define a rectangle area
teleport[1].z1=886		--NOTE: x1 & z1 should be the smaller coordinate numbers because corner order is used to
teleport[1].name = "left tunnel"
teleport[1].entrydir = "right" 	-- left | right 	TODO: up | down | any | exitonly
teleport[1].exitid = 2  	  	--units entering this teleport will teleport to the teleport with this id
teleport[1].exitdir = "left"	--units that exit here, will move to the middle of this side of the rectangle area

teleport[2] = {}
teleport[2].x1=8521
teleport[2].z1=886
teleport[2].name = "right tunnel"
teleport[2].entrydir = "left"
teleport[2].exitid = 1
teleport[2].exitdir = "right"

teleport[3] = {}
teleport[3].x1=4888		--4 corners to define a rectangle area
teleport[3].z1=3074		--NOTE: x1 & z1 should be the smaller coordinate numbers because corner order is used to
teleport[3].name = "left tunnel"
teleport[3].entrydir = "right" 	-- left | right 	TODO: up | down | any | exitonly
teleport[3].exitid = 4  	  	--units entering this teleport will teleport to the teleport with this id
teleport[3].exitdir = "left"	--units that exit here, will move to the middle of this side of the rectangle area

teleport[4] = {}
teleport[4].x1=7370
teleport[4].z1=3074
teleport[4].name = "right tunnel"
teleport[4].entrydir = "left"
teleport[4].exitid = 3
teleport[4].exitdir = "right"

teleport[5] = {}
teleport[5].x1=3690		--4 corners to define a rectangle area
teleport[5].z1=5270		--NOTE: x1 & z1 should be the smaller coordinate numbers because corner order is used to
teleport[5].name = "left tunnel"
teleport[5].entrydir = "right" 	-- left | right 	TODO: up | down | any | exitonly
teleport[5].exitid = 6  	  	--units entering this teleport will teleport to the teleport with this id
teleport[5].exitdir = "left"	--units that exit here, will move to the middle of this side of the rectangle area

teleport[6] = {}
teleport[6].x1=8520
teleport[6].z1=5270
teleport[6].name = "right tunnel"
teleport[6].entrydir = "left"
teleport[6].exitid = 5
teleport[6].exitdir = "right"

teleport[7] = {}
teleport[7].x1=1968		--4 corners to define a rectangle area
teleport[7].z1=3134		--NOTE: x1 & z1 should be the smaller coordinate numbers because corner order is used to
teleport[7].name = "left tunnel"
teleport[7].entrydir = "right" 	-- left | right 	TODO: up | down | any | exitonly
teleport[7].exitid = 8  	  	--units entering this teleport will teleport to the teleport with this id
teleport[7].exitdir = "left"	--units that exit here, will move to the middle of this side of the rectangle area

teleport[8] = {}
teleport[8].x1=10284
teleport[8].z1=3133
teleport[8].name = "right tunnel"
teleport[8].entrydir = "left"
teleport[8].exitid = 7
teleport[8].exitdir = "right"

teleport[9] = {}
teleport[9].x1=2045		--4 corners to define a rectangle area
teleport[9].z1=1970		--NOTE: x1 & z1 should be the smaller coordinate numbers because corner order is used to
teleport[9].name = "left tunnel"
teleport[9].entrydir = "right" 	-- left | right 	TODO: up | down | any | exitonly
teleport[9].exitid = 10  	  	--units entering this teleport will teleport to the teleport with this id
teleport[9].exitdir = "left"	--units that exit here, will move to the middle of this side of the rectangle area

teleport[10] = {}
teleport[10].x1=10195
teleport[10].z1=1970
teleport[10].name = "right tunnel"
teleport[10].entrydir = "left"
teleport[10].exitid = 9
teleport[10].exitdir = "right"

teleport[11] = {}
teleport[11].x1=2096		--4 corners to define a rectangle area
teleport[11].z1=4131		--NOTE: x1 & z1 should be the smaller coordinate numbers because corner order is used to
teleport[11].name = "left tunnel"
teleport[11].entrydir = "right" 	-- left | right 	TODO: up | down | any | exitonly
teleport[11].exitid = 12  	  	--units entering this teleport will teleport to the teleport with this id
teleport[11].exitdir = "left"	--units that exit here, will move to the middle of this side of the rectangle area

teleport[12] = {}
teleport[12].x1=10095
teleport[12].z1=4131
teleport[12].name = "right tunnel"
teleport[12].entrydir = "left"
teleport[12].exitid = 11
teleport[12].exitdir = "right"


-- when scrambling, odd should go into even and viceversa, to still allow map crossing
function scrambleTeleports()
	local numTeleports = #teleport
	local scrambleVectorLeft = {}
	local scrambleVectorRight = {}
	for count = 1, numTeleports, 1 do
		if count%2 == 1 then
			scrambleVectorLeft[#scrambleVectorLeft+1] = count+1
		else
			scrambleVectorRight[#scrambleVectorRight+1] = count-1
		end
	end
	for count = 1, numTeleports, 1 do
		if count%2 == 1 then -- separate odd & even
			local element = random( 1, #scrambleVectorLeft ) -- recalculate size everytime
			local exit = scrambleVectorLeft[element]
			teleport[count].exitid = exit
			if doScramble == "reciprocal" then
				teleport[exit].exitid = count -- keep teleports reciprocal
			end
			table.remove(scrambleVectorLeft,element)
		elseif doScramble == "shuffle" then -- if even number, and no reciprocal mode activated
			local element = random( 1, #scrambleVectorRight ) -- recalculate size everytime
			local exit = scrambleVectorRight[element]
			teleport[count].exitid = exit
			table.remove(scrambleVectorRight,element)
		end
	end
end

function notifyScramble()
	Echo("Teleport destinations have changed!") -- perhaps use something more visible?
end

function gadget:GameStart()
	-- save ground height were teleports are
	for ti = 1, #teleport, 1 do
		teleport[ti].groundheight = GetGroundHeight(teleport[ti].x1,teleport[ti].z1)
	end
end


function gadget:GameFrame(frame)
	
  --ugly hax dont look
	if frame == 1 then
		for ti = 1, #teleport, 1 do
			teleport[ti].groundheight = GetGroundHeight(teleport[ti].x1,teleport[ti].z1)
		end
	end
 --]] finished ugly hax
	
	if doScramble ~= "none" then
		if frame%scrambleinterval == 0 then
			scrambleTeleports()
			if frame > 0 then
				notifyScramble()
			end
		end
	end

	--Echo ("hey!")
	if (frame%30 == 0) then
		if (debug) then Echo ("checking for units in area") end
		for ti = 1, #teleport, 1 do
			units = GetUnitsInCylinder (teleport[ti].x1, teleport[ti].z1, teleportradius)
			--if (not units) then Spring.Echo ("keiner da") return end
			--if (#units >0) then Spring.Echo ("#units=" .. #units) end
			for i in pairs(units) do
				local unitID = units[i]
				if ValidUnitID (unitID) then
					local _,_,_,_,buildProgres = GetUnitHealth(unitID)
					if buildProgres and buildProgres == 1 then
						local _,basey = GetUnitBasePosition(unitID)
						if basey and ((basey - teleport[ti].groundheight) < 15) then -- small height marging, we're checking the unit bottom anyway to exclude flying units
							local dx, dy,dz = GetUnitDirection (unitID)
							local ed = teleport[ti].entrydir
							if (ed == "exit")
								then
								else
								if (ed == "left" and dx < 0)
								or (ed == "right" and dx > 0)
									then
									if (debug) then marker_on_unit (unitID, "in area " .. teleport[ti].name) end
									teleport_unit_in_area (unitID, teleport[ti].exitid)
								end
							end
						end
					end
				end
			end
		end
	end
	if frame%(16) == 0 then -- every slowupdate, restore the ground of 1 teleport
		local ti = (frame/16)%(#teleport) +1
		local x1 = teleport[ti].x1-teleportradius/2
		local z1 = teleport[ti].z1-teleportradius/2
		local x2 = x1 + teleportradius
		local z2 = z1 + teleportradius
		RevertHeightMap(x1,z1,x2,z2,1)
	end
	if ((frame+1)%2 == 0) then
		for ui in pairs(movecrtlunits) do
			--Echo ("unmoving: " .. ui)
			--Spring.MoveCtrl.Disable(ui)
			movecrtlunits[ui] = nil
			--GiveOrderToUnit(ui, CMD.MOVE , {2000, teleportmove, 2000  }, {})

			--local unitCommands = GetUnitCommands(ui)   --does not work
			--if #unitCommands > 0 then
				--GiveOrderToUnit(ui, unitCommands)
			--end
		end
	end
end

function gadget:UnitFinished(unitID, unitDefID, unitTeam)
	if UnitDefs[unitDefID]["canMove"] == 0 then -- directly move only static buildings
		return
	end
	basex,basey,basez = GetUnitBasePosition(unitID)
	local relativex
	local relativez
	local relativey
	local teleportID
	for ti = 1, #teleport, 1 do
		local x1 = teleport[ti].x1
		local z1 = teleport[ti].z1
		if ((basex-x1)^2 + (basez-z1)^2) < teleportradiussquared then
			relativex = basex - x1
			relativey = basey - GetGroundHeight(basex,basez)
			relativez = basez - z1
			teleportID = ti
 			break
 		end
	end
	if teleportID then
		local dest = teleport[teleportID].exitid
		local x1 = teleport[dest].x1
		local z1 = teleport[dest].z1
		local unitx = x1 + relativex
		local unitz = z1 + relativez
		local unity = GetGroundHeight(unitx,unitz) + relativey
		SetUnitPosition(unitID, unitx,unity,unitz)
	end
end

function gadget:Initialize()
	if (not debug) then return end
	Echo ("tp_wraparoundmap: yeaah here i am!")
	for i = 1, #teleport, 1 do
		Echo ("marking teleport #" .. i)
		MarkerAddPoint (teleport[i].x1,teleportmove, teleport[i].z1, teleport[i].name)
		MarkerAddPoint (teleport[i].x1,teleportmove, teleport[i].z2, teleport[i].name)
	end
end


function teleport_unit_in_area (ui, ti)
	local x = random (teleport[ti].x1-100, teleport[ti].x1+100)
	local z = random (teleport[ti].z1-100, teleport[ti].z1+100)
	Spring.MoveCtrl.Enable(ui)
	Spring.MoveCtrl.SetPosition (ui,x,teleportmove,z)
	Spring.MoveCtrl.Disable(ui)
	movecrtlunits [ui] = true
	local mtx, mtz	--movetartget (x,z) where should the unit move to after exiting the teleport?
	local ed = teleport[ti].exitdir
	if (ed=="left") then mtx = teleport[ti].x1-exitmoveaway mtz = (teleport[ti].z1+teleport[ti].z1)/2 end
	if (ed=="right") then mtx = teleport[ti].x1+exitmoveaway mtz = (teleport[ti].z1+teleport[ti].z1)/2 end
	if (debug) then MarkerAddPoint (mtx,teleportmove,mtz, "move target:" .. mtx ..":"..mtz) end
	GiveOrderToUnit(ui, CMD.MOVE , {mtx, teleportmove, mtz}, {})
end

function marker_on_unit (_uID, _text)
	if (not debug) then return end
	if (markers == false) then return end
	if (_uID == nil) then return end
	if (_text == nil) then return end
	local x,y,z=GetUnitPosition (_uID)
	if (x == nil or y == nil or z == nil) then return end
	MarkerAddPoint (x,y,z, _text .. "id:" .. _uID)
end

