----------------------------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------------------------

--  File:    MapOptions.lua
--  Description:   Custom MapOptions file that makes possible to set up variable options before game starts, like ModOptions.lua
--  Author:  SirArtturi, Lurker, Smoth


----------------------------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------------------------

--[[  OPTION DEFINITION FILES

  CustomMapOptions.lua
  - belongs in the map archive
  - options that get used by LuaGaia

  CustomModOptions.lua
  - belongs in the mod archive
  - options that get used by LuaRules

--]]

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

--  NOTES:
--  - using an enumerated table lets you specify the options order

--
--  These keywords must be lowercase for LuaParser to read them.
--
--  key:      the string used in the script.txt
--  name:     the displayed name
--  desc:     the description (could be used as a tooltip)
--  type:     the option type
--  def:      the default value
--  min:      minimum value for number options
--  max:      maximum value for number options
--  step:     quantization step, aligned to the def value
--  maxlen:   the maximum string length for string options
--  items:    array of item strings for list options
--  scope:    'all', 'player', 'team', 'allyteam'      <<< not supported yet >>>
--

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local options = {

	{
		key = 'Atmosphere',
		name = 'Atmosphere settings',
		desc = 'Weather and time',
		type = 'section',
	},
	 --{
		-- key="dry",
		 --name="Wetness",
		 --desc="How wet do you like it?",
		 --type="list",
		 --def="0",
		 --section= 'Terrain',
		 --items = {
		--	 { key = "0", name = "Wet", desc = "a river runs through"  },
		--	 { key = "1", name = "Dry", desc = "no water" }
		-- },
	 --},
		
	{
		 key = 'ma_scrambleteleports',
		 name = 'Scramble Teleports?',
		 desc = "possible options: none (no scramble), reciprocal (a<->b matching), shuffle ( a->b, b->c, c->a )",
		 type="list",
		 def="none",
		 section= 'Terrain',
		 items = {
			 { key = "none", name = "None", desc = "No scramble"  },
			 { key = "reciprocal", name = "Reciprocal", desc = "a<->b matching" },
			 { key = "shuffle", name = "Shuffle", desc = "a->b, b->c, c->a"  }
		 },
	},	
	{
		key="alt",
		name="Time of day",
		desc="Night or day?",
		type="list",
		def="0",
		section= 'Atmosphere',
		items = {
			{ key = "0", name = "Day", desc = "Day Time"  },
			{ key = "1", name = "Night", desc = "Night Time" }
		},
	},
	
	{
		key = 'fog',
		name = 'want fog?',
		desc = "enable a dense fog",
		type = 'bool',
		section = 'Atmosphere',
		def = false,
	},

}

return options